# pylint: disable=no-method-argument, no-self-argument, too-many-function-args
# pylint: disable=invalid-sequence-index, missing-kwoa, unsubscriptable-object
# pylint: disable=no-value-for-parameter, invalid-unary-operand-type
# pylint: disable=pointless-string-statement, pointless-statement
# pylint: disable=C0328
import functools
import traceback
import types
import unittest

from function_pipe.core import function_pipe as fpn  # pylint: disable=E0401


class TestUnit(unittest.TestCase):
    def test_basic_expressions_a(self) -> None:
        class TestInput(fpn.PipeNodeInput):
            def get_origin(self):
                return 32

        @fpn.pipe_node
        def proc_a(**kwargs):
            pni = kwargs[fpn.PN_INPUT]
            post = pni.get_origin()
            return post

        @fpn.pipe_node_factory
        def proc_b(scalar, **kwargs):
            pred = kwargs[fpn.PREDECESSOR_RETURN]
            post = pred * scalar
            return post

        @fpn.pipe_node
        def proc_c(**kwargs):
            pred = kwargs[fpn.PREDECESSOR_RETURN]
            post = pred / 10000
            return post

        @fpn.pipe_node
        def proc_d(**kwargs):
            pred = kwargs[fpn.PREDECESSOR_RETURN]
            return pred

        a = proc_a | proc_b(scalar=3) | proc_d | proc_b(scalar=0.5) | proc_c
        b = proc_a * 30
        c = proc_a | proc_b(scalar=b)
        d = proc_a | -proc_b(3)

        f = a + b

        pni = TestInput()

        post = f(pn_input=pni)
        self.assertEqual(post, 960.0048)

        post = a(pn_input=pni)
        self.assertEqual(post, 0.0048)

        post = b(pn_input=pni)
        self.assertEqual(post, 960)

        post = c(pn_input=pni)
        self.assertEqual(post, 30720)

        post = d(pn_input=pni)
        self.assertEqual(post, -96)

        # proc a can reside in non first position

        f = proc_a | proc_a | proc_a
        post = f(pn_input=pni)
        self.assertEqual(post, 32)

        f = proc_a | proc_a | proc_a
        post = f(pn_input=pni)
        self.assertEqual(post, 32)

    def test_basic_expressions_b(self) -> None:
        @fpn.pipe_node
        def foo1(**kwargs):
            return 12

        @fpn.pipe_node
        def bar1(**kwargs):
            return 13

        self.assertEqual(12, (bar1 | foo1)[None])
        self.assertEqual(13, (foo1 | bar1)[None])

    def test_basic_expressions_c(self) -> None:
        @fpn.pipe_node(fpn.PN_INPUT)
        def foo2(pni):
            return pni

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def bar2(pni, prev):
            return pni + prev

        self.assertEqual(27 * 2, (foo2 | bar2)[27])

    def test_basic_expressions_d(self) -> None:
        @fpn.pipe_node(fpn.PN_INPUT)
        def foo3(pni):
            return pni

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def bar3(pni, prev):
            return pni + prev

        self.assertEqual(27 * 2, (foo3 | bar3)[27])

    def test_basic_expressions_e(self) -> None:
        @fpn.pipe_node_factory
        def foo4(arg, **kwargs):
            return 12 + arg

        @fpn.pipe_node_factory
        def bar4(arg, **kwargs):
            return 13 - arg

        self.assertEqual(12 + 7, (bar4(6) | foo4(7))[None])
        self.assertEqual(13 - 6, (foo4(7) | bar4(6))[None])

    def test_basic_expressions_f(self) -> None:
        @fpn.pipe_node_factory(fpn.PN_INPUT)
        def foo5(pni, arg):
            return pni + arg

        @fpn.pipe_node_factory(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def bar5(pni, prev, arg):
            return pni + prev - arg

        self.assertEqual(27 * 2 + 8 - 2, (foo5(8) | bar5(2))[27])

    def test_basic_expressions_g(self) -> None:
        @fpn.pipe_node_factory(fpn.PN_INPUT)
        def foo6(pni, arg):
            return pni + arg

        @fpn.pipe_node_factory(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def bar6(pni, prev, arg):
            return pni + prev - arg

        self.assertEqual(27 * 2 + 8 - 2, (foo6(8) | bar6(2))[27])

    def test_methods_defined_on_classes(self) -> None:

        uno_mismo_pipe_node = functools.partial(fpn.pipe_node, self_keyword="uno_mismo")
        uno_mismo_pipe_node_factory = functools.partial(
            fpn.pipe_node_factory, self_keyword="uno_mismo"
        )

        class C:
            STATE = "C class state"

            def __init__(self, state) -> None:
                self.state = f"C {state}"

            # classmethods

            @fpn.classmethod_pipe_node
            def classmethod_node(cls, **kwargs):
                return cls.STATE

            @fpn.classmethod_pipe_node()
            def classmethod_node_bind_empty(cls):
                return "empty", cls.STATE

            @fpn.classmethod_pipe_node(fpn.PN_INPUT)
            def classmethod_node_bind(cls, pni):
                return pni, cls.STATE

            @fpn.classmethod_pipe_node_factory
            def classmethod_node_factory(cls, factory_val, **kwargs):
                return f"{cls.__name__} factory_val='{factory_val}'"

            @fpn.classmethod_pipe_node_factory()
            def classmethod_node_factory_bind_empty(cls, factory_val):
                return "empty", f"{cls.__name__} factory_val='{factory_val}'"

            @fpn.classmethod_pipe_node_factory(fpn.PN_INPUT)
            def classmethod_node_factory_bind(cls, pni, factory_val):
                return pni, f"{cls.__name__} factory_val='{factory_val}'"

            # staticmethods

            @fpn.staticmethod_pipe_node
            def staticmethod_node(**kwargs):
                return "STATIC"

            @fpn.staticmethod_pipe_node()
            def staticmethod_node_bind_empty():
                return "empty", "STATIC"

            @fpn.staticmethod_pipe_node(fpn.PN_INPUT)
            def staticmethod_node_bind(pni):
                return pni, "STATIC"

            @fpn.staticmethod_pipe_node_factory
            def staticmethod_node_factory(factory_val, **kwargs):
                return f"STATIC factory_val='{factory_val}'"

            @fpn.staticmethod_pipe_node_factory()
            def staticmethod_node_factory_bind_empty(factory_val):
                return "empty", f"STATIC factory_val='{factory_val}'"

            @fpn.staticmethod_pipe_node_factory(fpn.PN_INPUT)
            def staticmethod_node_factory_bind(pni, factory_val):
                return pni, f"STATIC factory_val='{factory_val}'"

            # namespace methods. These should fail if called from an instance.

            @fpn.pipe_node
            def namespace_node(**kwargs):
                return "NAMESPACE"

            @fpn.pipe_node()
            def namespace_node_bind_empty():
                return "empty", "NAMESPACE"

            @fpn.pipe_node(fpn.PN_INPUT)
            def namespace_node_bind(pni):
                return pni, "NAMESPACE"

            @fpn.pipe_node_factory
            def namespace_node_factory(factory_val, **kwargs):
                return f"NAMESPACE factory_val='{factory_val}'"

            @fpn.pipe_node_factory()
            def namespace_node_factory_bind_empty(factory_val):
                return "empty", f"NAMESPACE factory_val='{factory_val}'"

            @fpn.pipe_node_factory(fpn.PN_INPUT)
            def namespace_node_factory_bind(pni, factory_val):
                return pni, f"NAMESPACE factory_val='{factory_val}'"

            # self methods

            @fpn.pipe_node
            def self_node(self, **kwargs):
                return self.state

            @fpn.pipe_node()
            def self_node_bind_empty(self):
                return "empty", self.state

            @fpn.pipe_node(fpn.PN_INPUT)
            def self_node_bind(self, pni):
                return pni, self.state

            @fpn.pipe_node_factory
            def self_node_factory(self, factory_val, **kwargs):
                return f"SELF factory_val='{factory_val}'"

            @fpn.pipe_node_factory()
            def self_node_factory_bind_empty(self, factory_val):
                return "empty", f"SELF factory_val='{factory_val}'"

            @fpn.pipe_node_factory(fpn.PN_INPUT)
            def self_node_factory_bind(self, pni, factory_val):
                return pni, f"SELF factory_val='{factory_val}'"

            # self with non-standard first argument_name methods

            @uno_mismo_pipe_node
            def uno_mismo_node(uno_mismo, **kwargs):
                return uno_mismo.state

            @uno_mismo_pipe_node()
            def uno_mismo_node_bind_empty(uno_mismo):
                return "empty", uno_mismo.state

            @uno_mismo_pipe_node(fpn.PN_INPUT)
            def uno_mismo_node_bind(uno_mismo, pni):
                return pni, uno_mismo.state

            @uno_mismo_pipe_node_factory
            def uno_mismo_node_factory(uno_mismo, factory_val, **kwargs):
                return f"uno_mismo factory_val='{factory_val}'"

            @uno_mismo_pipe_node_factory()
            def uno_mismo_node_factory_bind_empty(uno_mismo, factory_val):
                return "empty", f"uno_mismo factory_val='{factory_val}'"

            @uno_mismo_pipe_node_factory(fpn.PN_INPUT)
            def uno_mismo_node_factory_bind(uno_mismo, pni, factory_val):
                return pni, f"uno_mismo factory_val='{factory_val}'"

        class D(C):
            STATE = "D class state"

            def __init__(self, state) -> None:
                self.state = f"D {state}"

        d = D("self state")
        pni = "PNI"

        # fmt: off
        self.assertEqual(d.classmethod_node[pni], "D class state")
        self.assertEqual(D.classmethod_node[pni], "D class state")
        self.assertEqual(C.classmethod_node[pni], "C class state")
        self.assertEqual(d.classmethod_node_bind_empty[pni], ("empty", "D class state"))
        self.assertEqual(D.classmethod_node_bind_empty[pni], ("empty", "D class state"))
        self.assertEqual(C.classmethod_node_bind_empty[pni], ("empty", "C class state"))
        self.assertEqual(d.classmethod_node_bind[pni], (pni, "D class state"))
        self.assertEqual(D.classmethod_node_bind[pni], (pni, "D class state"))
        self.assertEqual(C.classmethod_node_bind[pni], (pni, "C class state"))
        self.assertEqual(d.classmethod_node_factory("val")[pni], "D factory_val='val'")
        self.assertEqual(D.classmethod_node_factory("val")[pni], "D factory_val='val'")
        self.assertEqual(C.classmethod_node_factory("val")[pni], "C factory_val='val'")
        self.assertEqual(d.classmethod_node_factory_bind_empty("val")[pni], ("empty", "D factory_val='val'"))
        self.assertEqual(D.classmethod_node_factory_bind_empty("val")[pni], ("empty", "D factory_val='val'"))
        self.assertEqual(C.classmethod_node_factory_bind_empty("val")[pni], ("empty", "C factory_val='val'"))
        self.assertEqual(d.classmethod_node_factory_bind("val")[pni], (pni, "D factory_val='val'"))
        self.assertEqual(D.classmethod_node_factory_bind("val")[pni], (pni, "D factory_val='val'"))
        self.assertEqual(C.classmethod_node_factory_bind("val")[pni], (pni, "C factory_val='val'"))

        self.assertEqual(d.staticmethod_node[pni], "STATIC")
        self.assertEqual(D.staticmethod_node[pni], "STATIC")
        self.assertEqual(C.staticmethod_node[pni], "STATIC")
        self.assertEqual(d.staticmethod_node_bind_empty[pni], ("empty", "STATIC"))
        self.assertEqual(D.staticmethod_node_bind_empty[pni], ("empty", "STATIC"))
        self.assertEqual(C.staticmethod_node_bind_empty[pni], ("empty", "STATIC"))
        self.assertEqual(d.staticmethod_node_bind[pni], (pni, "STATIC"))
        self.assertEqual(D.staticmethod_node_bind[pni], (pni, "STATIC"))
        self.assertEqual(C.staticmethod_node_bind[pni], (pni, "STATIC"))
        self.assertEqual(d.staticmethod_node_factory("val")[pni], "STATIC factory_val='val'")
        self.assertEqual(D.staticmethod_node_factory("val")[pni], "STATIC factory_val='val'")
        self.assertEqual(C.staticmethod_node_factory("val")[pni], "STATIC factory_val='val'")
        self.assertEqual(d.staticmethod_node_factory_bind_empty("val")[pni], ("empty", "STATIC factory_val='val'"))
        self.assertEqual(D.staticmethod_node_factory_bind_empty("val")[pni], ("empty", "STATIC factory_val='val'"))
        self.assertEqual(C.staticmethod_node_factory_bind_empty("val")[pni], ("empty", "STATIC factory_val='val'"))
        self.assertEqual(d.staticmethod_node_factory_bind("val")[pni], (pni, "STATIC factory_val='val'"))
        self.assertEqual(D.staticmethod_node_factory_bind("val")[pni], (pni, "STATIC factory_val='val'"))
        self.assertEqual(C.staticmethod_node_factory_bind("val")[pni], (pni, "STATIC factory_val='val'"))

        self.assertEqual(d.namespace_node[pni], "NAMESPACE")
        self.assertEqual(D.namespace_node[pni], "NAMESPACE")
        self.assertEqual(C.namespace_node[pni], "NAMESPACE")

        self.assertEqual(d.namespace_node_bind[pni], (pni, "NAMESPACE"))
        self.assertEqual(D.namespace_node_bind[pni], (pni, "NAMESPACE"))
        self.assertEqual(C.namespace_node_bind[pni], (pni, "NAMESPACE"))
        self.assertEqual(d.namespace_node_bind_empty[pni], ("empty", "NAMESPACE"))
        self.assertEqual(D.namespace_node_bind_empty[pni], ("empty", "NAMESPACE"))
        self.assertEqual(C.namespace_node_bind_empty[pni], ("empty", "NAMESPACE"))
        self.assertEqual(d.namespace_node_factory("val")[pni], "NAMESPACE factory_val='val'")
        self.assertEqual(D.namespace_node_factory("val")[pni], "NAMESPACE factory_val='val'")
        self.assertEqual(C.namespace_node_factory("val")[pni], "NAMESPACE factory_val='val'")
        self.assertEqual(d.namespace_node_factory_bind_empty("val")[pni], ("empty", "NAMESPACE factory_val='val'"))
        self.assertEqual(D.namespace_node_factory_bind_empty("val")[pni], ("empty", "NAMESPACE factory_val='val'"))
        self.assertEqual(C.namespace_node_factory_bind_empty("val")[pni], ("empty", "NAMESPACE factory_val='val'"))
        self.assertEqual(d.namespace_node_factory_bind("val")[pni], (pni, "NAMESPACE factory_val='val'"))
        self.assertEqual(D.namespace_node_factory_bind("val")[pni], (pni, "NAMESPACE factory_val='val'"))
        self.assertEqual(C.namespace_node_factory_bind("val")[pni], (pni, "NAMESPACE factory_val='val'"))

        self.assertEqual(d.self_node[pni], "D self state")
        self.assertEqual(d.self_node_bind_empty[pni], ("empty", "D self state"))
        self.assertEqual(d.self_node_bind[pni], (pni, "D self state"))
        self.assertEqual(d.self_node_factory("val")[pni], "SELF factory_val='val'")
        self.assertEqual(d.self_node_factory_bind_empty("val")[pni], ("empty", "SELF factory_val='val'"))
        self.assertEqual(d.self_node_factory_bind("val")[pni], (pni, "SELF factory_val='val'"))

        self.assertEqual(d.uno_mismo_node[pni], "D self state")
        self.assertEqual(d.uno_mismo_node_bind_empty[pni], ("empty", "D self state"))
        self.assertEqual(d.uno_mismo_node_bind[pni], (pni, "D self state"))
        self.assertEqual(d.uno_mismo_node_factory("val")[pni], "uno_mismo factory_val='val'")
        self.assertEqual(d.uno_mismo_node_factory_bind_empty("val")[pni], ("empty", "uno_mismo factory_val='val'"))
        self.assertEqual(d.uno_mismo_node_factory_bind("val")[pni], (pni, "uno_mismo factory_val='val'"))
        # fmt: on

        """Custom ValueError indicating an unbound self-method was called without an instance"""
        with self.assertRaises(ValueError):
            D.self_node[pni]

        with self.assertRaises(ValueError):
            C.self_node[pni]

        """Normal errors about a missing required positional arg, which makes sense, since self wasn't passed in"""
        with self.assertRaises(TypeError):
            D.self_node_bind[pni]

        with self.assertRaises(TypeError):
            C.self_node_bind[pni]

        with self.assertRaises(TypeError):
            D.self_node_factory("val")[pni]

        with self.assertRaises(TypeError):
            C.self_node_factory("val")[pni]

        with self.assertRaises(TypeError):
            D.self_node_factory_bind("val")[pni]

        with self.assertRaises(TypeError):
            C.self_node_factory_bind("val")[pni]

        """Errors about re-providing bound positional arguments like self/cls"""
        with self.assertRaises(ValueError):
            D.classmethod_node(cls="re-provided", **{fpn.PN_INPUT: pni})

        with self.assertRaises(TypeError):
            d.self_node(  # pylint: disable=redundant-keyword-arg)
                self="re-provided", **{fpn.PN_INPUT: pni}
            )

        with self.assertRaises(ValueError):
            d.uno_mismo_node(  # pylint: disable=redundant-keyword-arg)
                uno_mismo="re-provided", **{fpn.PN_INPUT: pni}
            )

    def test_bound_and_unbound_pipe_nodes(self) -> None:
        @fpn.pipe_node
        def pn_unbound(**kwargs):
            return kwargs[fpn.PN_INPUT]

        @fpn.pipe_node(fpn.PN_INPUT)
        def pn_bound(pni):
            return pni

        @fpn.pipe_node_factory
        def pnf_unbound(factory_val, **kwargs):
            return kwargs[fpn.PN_INPUT], factory_val

        @fpn.pipe_node_factory(fpn.PN_INPUT)
        def pnf_bound(pni, factory_val):
            return pni, factory_val

        pni = "PNI"

        self.assertEqual(pn_unbound[pni], pn_bound[pni])
        self.assertEqual(pnf_unbound("fval")[pni], pnf_bound("fval")[pni])

    def test_complex_pipeline(self) -> None:
        @fpn.pipe_node_factory(fpn.PREDECESSOR_RETURN)
        def multiply(prev, val):
            return prev * val

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def add_pni(pni, prev):
            return prev + pni.value

        @fpn.pipe_node
        def init(**kwargs):
            return 12

        @fpn.pipe_node_factory
        def func(arg1, arg2, **kwargs):
            return (kwargs[fpn.PN_INPUT].value * arg1) + (
                kwargs[fpn.PREDECESSOR_RETURN] / arg2
            )

        class Operations:
            DELTA = 0.23

            def __init__(self, factor):
                self.factor = factor

            @fpn.classmethod_pipe_node
            def add_delta(cls, **kwargs):
                return cls.DELTA + kwargs[fpn.PREDECESSOR_RETURN]

            @fpn.staticmethod_pipe_node_factory(fpn.PREDECESSOR_RETURN)
            def bound_prev(prev, lower, upper):
                return max(min(prev, lower), upper)

            @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
            def adj_by_factor(self, pni, prev):
                return self.factor * (pni.value - prev)

        op = Operations(23.717)

        expr = (
            init
            | add_pni
            | fpn.store("A")
            | multiply(-8)
            | func(13, 7)
            | fpn.store("B")
            | op.add_delta
            | Operations.add_delta
            | op.bound_prev(0, 100)
            | fpn.store("C")
            | multiply(130.2)
            | Operations.bound_prev(100, 200)
            | op.adj_by_factor
            | fpn.store("D")
            | fpn.recall("A")
            | op.adj_by_factor
            | fpn.recall("B")
            | op.adj_by_factor
            | fpn.recall("C")
            | op.adj_by_factor
            | add_pni
        )

        class PNI(fpn.PipeNodeInput):
            def __init__(self, value):
                super().__init__()
                self.value = value

        pni = PNI(-89)
        post = expr[pni]

        # Manually calculate
        expected = 12
        expected += pni.value
        A = expected
        expected *= -8
        expected = (pni.value * 13) + (expected / 7)
        B = expected
        expected += op.DELTA
        expected += Operations.DELTA
        expected = max(min(expected, 0), 100)
        C = expected
        expected *= 130.2
        expected = max(min(expected, 100), 200)
        expected = op.factor * (pni.value - expected)
        D = expected
        expected = op.factor * (pni.value - C)
        expected += pni.value

        self.assertEqual(-4571.513, expected)
        self.assertEqual(-4571.513, post)

        self.assertEqual(A, pni._store["A"])
        self.assertEqual(B, pni._store["B"])
        self.assertEqual(C, pni._store["C"])
        self.assertEqual(D, pni._store["D"])

        self.assertEqual(pni.store_items, dict(A=A, B=B, C=C, D=D).items())

    def test_core_decorator(self) -> None:
        # Modify the core_decorators
        found_f = set()

        PREFIX = "TestUnit.test_core_decorator.<locals>."

        def decorator(f):
            found_f.add(repr(f).replace(PREFIX, "").split()[1])
            return f

        pipe_node = functools.partial(fpn.pipe_node, core_decorator=decorator)
        pipe_node_factory = functools.partial(
            fpn.pipe_node_factory, core_decorator=decorator
        )
        classmethod_pipe_node = functools.partial(
            fpn.classmethod_pipe_node, core_decorator=decorator
        )
        staticmethod_pipe_node_factory = functools.partial(
            fpn.staticmethod_pipe_node_factory, core_decorator=decorator
        )

        @pipe_node_factory(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def store(pni, ret_val, label):
            pni.store(label, ret_val)
            return ret_val

        @pipe_node_factory(fpn.PN_INPUT)
        def recall(pni, label):
            return pni.recall(label)

        # ---------------------------------------------------------------------

        @pipe_node_factory(fpn.PREDECESSOR_RETURN)
        def multiply(prev, val):
            return prev * val

        @pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def add_pni(pni, prev):
            return prev + pni.value

        @pipe_node
        def init(**kwargs):
            return 12

        @pipe_node_factory
        def func(arg1, arg2, **kwargs):
            return (kwargs[fpn.PN_INPUT].value * arg1) + (
                kwargs[fpn.PREDECESSOR_RETURN] / arg2
            )

        class Operations:
            DELTA = 0.23

            def __init__(self, factor):
                self.factor = factor

            @classmethod_pipe_node
            def add_delta(cls, **kwargs):
                return cls.DELTA + kwargs[fpn.PREDECESSOR_RETURN]

            @staticmethod_pipe_node_factory(fpn.PREDECESSOR_RETURN)
            def bound_prev(prev, lower, upper):
                return max(min(prev, lower), upper)

            @pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
            def adj_by_factor(self, pni, prev):
                return self.factor * (pni.value - prev)

        op = Operations(23.717)

        expr = (
            init
            | add_pni
            | store("A")
            | multiply(-8)
            | func(13, 7)
            | store("B")
            | op.add_delta
            | Operations.add_delta
            | op.bound_prev(0, 100)
            | store("C")
            | multiply(130.2)
            | Operations.bound_prev(100, 200)
            | op.adj_by_factor
            | store("D")
            | recall("A")
            | op.adj_by_factor
            | recall("B")
            | op.adj_by_factor
            | recall("C")
            | op.adj_by_factor
            | add_pni
        )

        class PNI(fpn.PipeNodeInput):
            def __init__(self, value):
                super().__init__()
                self.value = value

        pni = PNI(-89)
        post = expr[pni]
        self.assertEqual(-4571.513, post)

        self.assertSetEqual(
            {
                "store",
                "recall",
                "init",
                "multiply",
                "func",
                "Operations.add_delta",
                "Operations.bound_prev",
                "Operations.adj_by_factor",
                "add_pni",
            },
            found_f,
        )

    def test_pn_repr_a(self) -> None:
        @fpn.pipe_node_factory()
        def factory(arg1, arg2, *, kwarg):
            pass

        self.assertEqual("<PNF: factory>", str(factory))
        self.assertEqual("<PNF: factory>", repr(factory))
        self.assertEqual("<PN: factory(1,2,kwarg=3)>", repr(factory(1, 2, kwarg=3)))
        self.assertEqual(
            "<PN: factory(1,2,kwarg=3) | factory(1,2,kwarg=3)>",
            repr(factory(1, 2, kwarg=3) | factory(1, 2, kwarg=3)),
        )
        self.assertEqual("<PN: factory(1,2,kwarg=3)>", str(factory(1, 2, kwarg=3)))
        self.assertEqual(
            "<PN: factory(1,2,kwarg=3) | factory(1,2,kwarg=3)>",
            str(factory(1, 2, kwarg=3) | factory(1, 2, kwarg=3)),
        )

        @fpn.pipe_node()
        def pn1():
            pass

        @fpn.pipe_node()
        def pn2():
            pass

        @fpn.pipe_node_factory()
        def pn3(arg1, *arg_list, kwarg, **kwargs):
            pass

        args = (1, 2, 3)
        kwargs = dict(kwarg=4, a=5, b=6)

        self.assertEqual(
            "<PN: pn1+pn2 | pn3(1,2,3,kwarg=4,a=5,b=6)>",
            repr((pn1 + pn2) | pn3(*args, **kwargs)),
        )
        self.assertEqual(
            "<PN: pn1 | pn2 | pn3(1,2,3,kwarg=4,a=5,b=6) | pn2 | pn1>",
            repr(pn1 | pn2 | pn3(*args, **kwargs) | pn2 | pn1),
        )
        self.assertEqual(
            "<PN: pn1+pn2 | pn3(1,2,3,kwarg=4,a=5,b=6)>",
            str((pn1 + pn2) | pn3(*args, **kwargs)),
        )
        self.assertEqual(
            "<PN: pn1 | pn2 | pn3(1,2,3,kwarg=4,a=5,b=6) | pn2 | pn1>",
            str(pn1 | pn2 | pn3(*args, **kwargs) | pn2 | pn1),
        )

    def test_pn_repr_b(self) -> None:
        @fpn.pipe_node()
        def a():
            return 1

        @fpn.pipe_node()
        def b():
            return 2

        @fpn.pipe_node()
        def c():
            return 3

        @fpn.pipe_node()
        def d():
            return 4

        @fpn.pipe_node()
        def e():
            return 5

        neg_a = -a
        abs_b = abs(b)
        inv_c = ~c

        self.assertEqual("<PN: -a>", repr(neg_a))
        self.assertEqual("<PN: abs(b)>", repr(abs_b))
        self.assertEqual("<PN: ~c>", repr(inv_c))

        add = a + b
        sub = a - b
        mul = a * b
        div = a / b

        self.assertEqual("<PN: a+b>", repr(add))
        self.assertEqual("<PN: a-b>", repr(sub))
        self.assertEqual("<PN: a*b>", repr(mul))
        self.assertEqual("<PN: a/b>", repr(div))

        add_mul_sub = add * sub
        sub_div_add = sub / add
        mul_add_div = mul + div
        div_sub_mul = div - mul
        add_pow_add = add**add

        self.assertEqual("<PN: (a+b)*(a-b)>", repr(add_mul_sub))
        self.assertEqual("<PN: (a-b)/(a+b)>", repr(sub_div_add))
        self.assertEqual("<PN: (a*b)+(a/b)>", repr(mul_add_div))
        self.assertEqual("<PN: (a/b)-(a*b)>", repr(div_sub_mul))
        self.assertEqual("<PN: (a+b)**(a+b)>", repr(add_pow_add))

        add_mul_sub_eq = add_mul_sub == c
        sub_div_add_gt = sub_div_add > d
        mul_add_div_lt = mul_add_div < e
        div_sub_mul_ge = div_sub_mul >= c
        add_pow_add_le = add_pow_add <= d
        add_pow_add_ne = add_pow_add != e

        self.assertEqual("<PN: ((a+b)*(a-b))==c>", repr(add_mul_sub_eq))
        self.assertEqual("<PN: ((a-b)/(a+b))>d>", repr(sub_div_add_gt))
        self.assertEqual("<PN: ((a*b)+(a/b))<e>", repr(mul_add_div_lt))
        self.assertEqual("<PN: ((a/b)-(a*b))>=c>", repr(div_sub_mul_ge))
        self.assertEqual("<PN: ((a+b)**(a+b))<=d>", repr(add_pow_add_le))
        self.assertEqual("<PN: ((a+b)**(a+b))!=e>", repr(add_pow_add_ne))

        complex_a = ((a <= ~b) >= c) + d - abs(e) * a ** ((b / -c) != (d == e))
        complex_b = (a <= b) >= (-c + (d - (e * a) ** b) / (c != abs(d)) == ~e)
        complex_c = ((-a <= abs(b)) >= (c + d)) - (e * a) ** (~b / c) != (d == e)
        complex_d = ((a <= b) >= c) + (~d - ((abs(-e) * a) ** b) / (c != d) == e)
        complex_e = abs(a) <= b >= c + ~d - e * a**b / c != -d == e

        self.assertEqual(
            "<PN: (((a<=~b)>=c)+d)-(abs(e)*(a**((b/-c)!=(d==e))))>", repr(complex_a)
        )
        self.assertEqual(
            "<PN: (a<=b)>=((-c+((d-((e*a)**b))/(c!=abs(d))))==~e)>", repr(complex_b)
        )
        self.assertEqual(
            "<PN: (((-a<=abs(b))>=(c+d))-((e*a)**(~b/c)))!=(d==e)>", repr(complex_c)
        )
        self.assertEqual(
            "<PN: ((a<=b)>=c)+((~d-((((abs(-e))*a)**b)/(c!=d)))==e)>", repr(complex_d)
        )
        self.assertEqual("<PN: -d==e>", repr(complex_e))

    def test_pn_repr_c(self) -> None:
        lambda_func = lambda x: x + 2
        fn = fpn.pipe_node(lambda_func)
        self.assertEqual("<PN: lambda_func = lambda x: x + 2>", repr(fn))
        self.assertEqual(
            "<PN: repr(fpn.pipe_node(lambda x: x + 2)),>",
            repr(fpn.pipe_node(lambda x: x + 2)),
        )  # Isn't really an easy way to parse the lambda expression alone.

    def test_cannot_store_twice(self) -> None:
        @fpn.pipe_node()
        def pn():
            return 12

        pni = fpn.PipeNodeInput()

        with self.assertRaises(KeyError):
            (pn | fpn.store("a") | fpn.store("a"))[pni]

    def test_call(self) -> None:
        calls = []

        @fpn.pipe_node(fpn.PN_INPUT)
        def pn1(pni):
            self.assertEqual("pni", pni)
            calls.append("pn1")
            return 13

        @fpn.pipe_node(fpn.PN_INPUT)
        def pn2(pni):
            self.assertEqual("pni", pni)
            calls.append("pn2")
            return 14

        @fpn.pipe_node(fpn.PN_INPUT)
        def pn3(pni):
            self.assertEqual("pni", pni)
            calls.append("pn3")
            return 15

        call = fpn.call(pn1, pn2, pn3)
        post = call["pni"]
        self.assertEqual(15, post)

        self.assertListEqual(["pn1", "pn2", "pn3"], calls)

    def test_invalid_operations(self) -> None:
        @fpn.pipe_node()
        def pn1():
            pass

        @fpn.pipe_node()
        def pn2():
            pass

        class MissingOp:
            pass

        with self.assertRaises(NotImplementedError):
            pn1 >> pn2

        with self.assertRaises(NotImplementedError):
            MissingOp() >> pn1

        with self.assertRaises(NotImplementedError):
            pn1 << pn2

        with self.assertRaises(NotImplementedError):
            MissingOp() << pn1

        with self.assertRaises(NotImplementedError):
            pn1.partial(1, 2, a=3, b=4)

    def test_pn_states(self) -> None:
        @fpn.pipe_node_factory()
        def factory(*args, **kwargs):
            pass

        @fpn.pipe_node()
        def pn():
            pass

        self.assertEqual(fpn.PipeNode.State.PROCESS, (pn | factory()).call_state)
        self.assertEqual(fpn.PipeNode.State.EXPRESSION, pn.call_state)
        self.assertEqual(fpn.PipeNode.State.FACTORY, factory.call_state)

    def test_predecessor_property(self) -> None:
        testable = {}

        @fpn.pipe_node()
        def init():
            pass

        @fpn.pipe_node(fpn.PREDECESSOR_PN)
        def pn1(prev_pn):
            testable["pn1"] = prev_pn

        @fpn.pipe_node(fpn.PREDECESSOR_PN)
        def pn2(prev_pn):
            testable["pn2"] = prev_pn

        expr = init | pn1 | pn2
        self.assertEqual({}, testable)
        self.assertIsNone(pn1.predecessor)
        self.assertIsNone(pn2.predecessor)

        expr[None]
        self.assertEqual({"pn1": pn1.predecessor, "pn2": pn2.predecessor}, testable)

    def test_ror(self) -> None:
        class MissingOR:
            def __init__(self, state):
                self.state = state

            def __call__(self, *args, **kwargs):
                print(args, kwargs)
                return self.state

        obj = MissingOR(37)

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN, fpn.PREDECESSOR_PN)
        def pn(pni, prev, prev_pn):
            self.assertIs(obj, prev_pn)
            self.assertEqual(37, prev)
            self.assertEqual("pni", pni)
            return 343

        expr = obj | pn
        post = expr["pni"]
        self.assertEqual(343, post)

    def test_invalid_factories(self) -> None:
        @fpn.pipe_node_factory
        def bad_factory_a(**kwargs):
            pass

        @fpn.pipe_node_factory()
        def bad_factory_b(pn_input):
            pass

        @fpn.pipe_node_factory()
        def good_factory(a, b, *, c, d=None):
            pass

        pn = fpn.pipe_node()(lambda: None)

        # Cannot use reserved kwargs!
        for kwarg in fpn.PIPE_NODE_KWARGS:
            with self.assertRaises(ValueError):
                bad_factory_a(**{kwarg: "something"})

            with self.assertRaises(ValueError):
                bad_factory_b(**{kwarg: "something"})

        # Cannot put a factory inside a pipeline!
        with self.assertRaises(ValueError):
            pn | good_factory

        # Calling a factory with an inadequate number of args/kwargs
        expr_should_fail = pn | good_factory(1, d=14)
        with self.assertRaises(TypeError):
            expr_should_fail[None]

        # Seeding factory with nothing, and then passing in reserved kwargs
        with self.assertRaises(ValueError):
            bad_factory_a()(pn_input=True, predecessor_return=True)

    def test_invalid_pn_call(self) -> None:
        @fpn.pipe_node()
        def pn_a():
            pass

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def pn_b(pni, prev):
            pass

        with self.assertRaises(ValueError):
            pn_a(1, 2, 3)

        with self.assertRaises(ValueError):
            pn_b(1, 2, 3)

        with self.assertRaises(ValueError):
            pn_a(kwarg=13)  # pylint: disable=unexpected-keyword-arg

        with self.assertRaises(ValueError):
            pn_b(kwarg=13)  # pylint: disable=unexpected-keyword-arg

    def test_compose(self) -> None:
        def square(x: int) -> int:
            return x**2

        def half(x: int) -> int:
            return x // 2

        functions = (square, half, half, square, half)

        fnA = fpn.compose(*functions)
        assert fnA(14) == square(half(half(square(half(14))))) == 144

        fnB = fpn.compose(*reversed(functions))
        assert fnB(14) == half(square(half(half(square(14))))) == 1200

    def test_fn_str_repr(self) -> None:
        def inc(x):
            return x + 1

        def square(x: int) -> int:
            return x**2

        def half(x: int) -> int:
            return x // 2

        fnA = fpn.compose(square, half, half, square, half)
        fnB = fpn.FunctionNode(inc)
        fnC = fnA >> fnB
        fnD = fnB + fnA

        self.assertEqual(str(fnA), repr(fnA))
        self.assertEqual(str(fnB), repr(fnB))
        self.assertEqual(str(fnC), repr(fnC))
        self.assertEqual(str(fnD), repr(fnD))

        self.assertEqual(str(fnA), "<FN: compose(half,square,half,half,square)>")
        self.assertEqual(str(fnB), "<FN: inc>")
        self.assertEqual(
            str(fnC), "<FN: compose(compose(half,square,half,half,square),inc)>"
        )
        self.assertEqual(str(fnD), "<FN: inc+(compose(half,square,half,half,square))>")

    def test_fn_partialling(self) -> None:
        def func(a, b, *, c, d=0):
            print(a, b, c, d)
            return a + b + c + d

        fnA = fpn.FunctionNode(func)
        self.assertEqual(fnA(1, 2, c=3, d=4), 10)

        fnB = fnA + fnA
        self.assertEqual(fnB(1, 2, c=3, d=4), 20)

        fnC = fnA >> fnB
        with self.assertRaises(TypeError):
            fnC(1, 2, c=3, d=4)

        fnD = fnA >> fnB.partial(20, c=30, d=40)
        self.assertEqual(fnD(1, 2, c=3, d=4), 200)

    def test_exception_with_cleaned_tb(self) -> None:
        class TestingError(RuntimeError):
            pass

        def func():
            raise TestingError("Testing")

        try:
            func()
        except TestingError as e:
            # We can not give this function arbitrary exceptions. It requires
            # exceptions caught from _inside_ function_pipe.py
            with self.assertRaises(AssertionError):
                fpn._exception_with_cleaned_tb(e)

        node = fpn.FunctionNode(func)

        try:
            node()
        except TestingError as e:
            frame_reprs = traceback.format_tb(e.__traceback__)

        self.assertEqual(len(frame_reprs), 3)
        self.assertIn("test_function_pipe.py", frame_reprs[0])
        self.assertIn("test_exception_with_cleaned_tb", frame_reprs[0])
        self.assertNotIn("test_function_pipe.py", frame_reprs[1])
        self.assertIn("function_pipe.py", frame_reprs[1])
        self.assertIn("_exception_with_cleaned_tb", frame_reprs[1])
        self.assertIn("test_function_pipe.py", frame_reprs[2])
        self.assertIn("TestingError", frame_reprs[2])

    def test_fn_operators_a(self) -> None:
        fn = fpn.FunctionNode(lambda x: x + 1)

        self.assertEqual((fn + 3)(1), 2 + 3)
        self.assertEqual((3 + fn)(1), 3 + 2)
        self.assertEqual((fn - 3)(1), 2 - 3)
        self.assertEqual((3 - fn)(1), 3 - 2)
        self.assertEqual((fn * 3)(1), 2 * 3)
        self.assertEqual((3 * fn)(1), 3 * 2)
        self.assertEqual((fn / 3)(1), 2 / 3)
        self.assertEqual((3 / fn)(1), 3 / 2)

    def test_fn_operators_b(self) -> None:
        half = fpn.FunctionNode(lambda x: x // 2)
        square = lambda x: x**2

        self.assertEqual((half >> square)(1), square(half(1)))
        self.assertEqual((square >> half)(1), half(square(1)))

        self.assertEqual((half << square)(1), half(square(1)))
        self.assertEqual((square << half)(1), square(half(1)))

        with self.assertRaises(NotImplementedError):
            half | square

        with self.assertRaises(NotImplementedError):
            square | half

    def test_is_unbound_self_method(self) -> None:
        def classmethod_func(cls):
            pass

        def staticmethod_func():
            pass

        class TestClass:
            def m_self(self):
                pass

            def m_uno_mismo(uno_mismo):
                pass

            @classmethod
            def m_classmethod_a(cls):
                pass

            @staticmethod
            def m_staticmethod_a():
                pass

            m_classmethod_b = classmethod(classmethod_func)
            m_staticmethod_b = staticmethod(staticmethod_func)

        m_classmethod_c = classmethod(classmethod_func)
        m_staticmethod_c = staticmethod(staticmethod_func)

        func_self = functools.partial(fpn._is_unbound_self_method, self_keyword="self")
        func_uno_mismo = functools.partial(
            fpn._is_unbound_self_method, self_keyword="uno_mismo"
        )

        self.assertTrue(func_self(TestClass.m_self))
        self.assertFalse(func_uno_mismo(TestClass.m_self))
        self.assertFalse(func_self(TestClass.m_uno_mismo))
        self.assertTrue(func_uno_mismo(TestClass.m_uno_mismo))

        self.assertFalse(func_self(TestClass.m_classmethod_a))
        self.assertFalse(func_uno_mismo(TestClass.m_classmethod_a))
        self.assertFalse(func_self(TestClass.m_staticmethod_a))
        self.assertFalse(func_uno_mismo(TestClass.m_staticmethod_a))

        self.assertFalse(func_self(TestClass.m_classmethod_b))
        self.assertFalse(func_uno_mismo(TestClass.m_classmethod_b))
        self.assertFalse(func_self(TestClass.m_staticmethod_b))
        self.assertFalse(func_uno_mismo(TestClass.m_staticmethod_b))

        self.assertFalse(func_self(m_classmethod_c))
        self.assertFalse(func_uno_mismo(m_classmethod_c))
        self.assertFalse(func_self(m_staticmethod_c))
        self.assertFalse(func_uno_mismo(m_staticmethod_c))

        self.assertFalse(func_self(functools.partial(classmethod_func, cls=None)))
        self.assertFalse(func_uno_mismo(functools.partial(classmethod_func, cls=None)))

    def test_unwrapping(self) -> None:
        @fpn.pipe_node
        def pn1(*args, **kwargs):
            return 1

        @fpn.pipe_node(fpn.PN_INPUT, fpn.PREDECESSOR_RETURN)
        def pn2(pni, prev):
            return pni + prev * 2

        @fpn.pipe_node_factory
        def pn3(a, *, b, c=None, **kwargs):
            return a + b - (c if c else 0) + kwargs[fpn.PREDECESSOR_RETURN]

        @fpn.pipe_node_factory(fpn.PREDECESSOR_RETURN)
        def pn4(prev, val):
            return prev**val

        def test_pn1(pn_or_expr):
            self.assertEqual(1, pn_or_expr.unwrap(1, 2, 3, 4, 5, 6))
            self.assertEqual(1, pn_or_expr.unwrap(kwargs_accept_anything=True))
            self.assertIsInstance(pn_or_expr.unwrap, types.FunctionType)
            self.assertFalse(isinstance(pn_or_expr.unwrap, fpn.FunctionNode))

        test_pn1(pn1)
        test_pn1(pn2 | pn1)
        test_pn1(pn2 | pn4() | pn1)

        def test_pn2(pn_or_expr) -> None:
            with self.assertRaises(KeyError):
                pn_or_expr.unwrap(1, 2)

            with self.assertRaises(KeyError):
                pn_or_expr.unwrap(pni=1, prev=2)

            self.assertEqual(16, pn_or_expr.unwrap(pn_input=6, predecessor_return=5))
            self.assertIsInstance(pn_or_expr.unwrap, types.FunctionType)
            self.assertFalse(isinstance(pn_or_expr.unwrap, fpn.FunctionNode))

        test_pn2(pn2)
        test_pn2(pn1 | pn2)
        test_pn2(pn1 | pn3() | pn2)

        def test_pn3(pn_or_expr) -> None:
            with self.assertRaises(KeyError):
                pn_or_expr.unwrap(1, b=2, c=8)  # Missing predecessor_return

            self.assertEqual(
                17.8, pn_or_expr.unwrap(1, b=-0.2, c=-3, predecessor_return=14)
            )
            self.assertIsInstance(pn_or_expr.unwrap, types.FunctionType)
            self.assertFalse(isinstance(pn_or_expr.unwrap, fpn.FunctionNode))

        test_pn3(pn3)
        test_pn3(pn1 | pn3())
        test_pn3(pn2 | pn4(1) | pn3())

        # pn4
        def test_pn4(pn_or_expr) -> None:
            with self.assertRaises(KeyError):
                pn_or_expr.unwrap(1)  # Missing predecessor_return

            with self.assertRaises(KeyError):
                pn_or_expr.unwrap(prev=1)  # Missing predecessor_return

            self.assertEqual(
                4.871658325766914, pn_or_expr.unwrap(0.6, predecessor_return=14)
            )
            self.assertIsInstance(pn_or_expr.unwrap, types.FunctionType)
            self.assertFalse(isinstance(pn_or_expr.unwrap, fpn.FunctionNode))

        test_pn4(pn4)
        test_pn4(pn2 | pn4())
        test_pn4(pn1 | pn3() | pn4())



# pylint: disable=C0328
"""
function_pipe.py

Copyright 2012-2022 Research Affiliates

Authors: Christopher Ariza, Max Moroz, Charles Burkland

Common usage:
import function_pipe as fpn
"""
import enum
import functools
import inspect
import re
import sys
import types
import typing as tp

# -------------------------------------------------------------------------------
# FunctionNode and utilities

FN = tp.TypeVar("FN", bound="FunctionNode")
PN = tp.TypeVar("PN", bound="PipeNode")
PNI = tp.TypeVar("PNI", bound="PipeNodeInput")
PipeNodeDescriptorT = tp.TypeVar("PipeNodeDescriptorT", bound="PipeNodeDescriptor")
KeyPostion = tp.Union[tp.Callable, str]
HandlerT = tp.Callable[[tp.Any], tp.Callable]


def compose(*funcs: tp.Callable) -> FN:
    """
    Given a list of functions, execute them from right to left, passing
    the returned value of the right f to the left f. Store the reduced function in a FunctionNode
    """
    # call right first, then left of each pair; each reduction retruns a function
    reducer = functools.reduce(
        lambda f, g: lambda *args, **kaargs: f(g(*args, **kaargs)), funcs
    )
    # args are reversed to show execution from right to left
    return FunctionNode(  # type: ignore
        reducer,
        doc_function=compose,
        doc_args=tuple(reversed(funcs)),
    )


def _wrap_unary(func: tp.Callable[[FN], FN]) -> tp.Callable[[FN], FN]:
    """
    Decorator for operator overloads. Given a higher order function that takes one args, wrap it in a FunctionNode function and provide documentation labels.
    """

    @functools.wraps(func)
    def unary(lhs: FN) -> FN:
        # wrapped function will prepare correct class, even if a constant
        cls = PipeNode if isinstance(lhs, PipeNode) else FunctionNode
        return cls(func(lhs), doc_function=func, doc_args=(lhs,))

    return unary


def _wrap_binary(func: tp.Callable[[FN, tp.Any], FN]) -> tp.Callable[[FN, tp.Any], FN]:
    """Decorator for operators. Given a higher order function that takes two args, wrap it in a FunctionNode function and provide documentation labels."""

    def binary(lhs: FN, rhs: tp.Any) -> FN:
        # wrapped function will prepare correct class, even if a constant
        cls = PipeNode if isinstance(lhs, PipeNode) else FunctionNode
        return cls(func(lhs, rhs), doc_function=func, doc_args=(lhs, rhs))

    return binary


def _from_this_module(tb: types.TracebackType) -> bool:
    return tb.tb_frame.f_code.co_filename == __file__


def _exception_with_cleaned_tb(original_exception: Exception) -> Exception:
    """
    Return back a new exception, where traceback is pointing to the first frame with code originating outside this module.
    """
    tb = original_exception.__traceback__
    assert tb is not None
    tb_next = None

    assert _from_this_module(tb)  # Sanity

    while True:
        tb_next = tb.tb_next

        # If `tb_next` is None, it means everything from the stack came from this module.
        # Use the original exception
        if tb_next is None:
            return original_exception

        # If `tb_next` originates from outside the module, it means we are done looking
        if not _from_this_module(tb_next):
            break

        # We are still observing frames from inside the module; keep looking
        tb = tb_next

    return original_exception.__class__(*original_exception.args).with_traceback(
        tb_next
    )


_BINARY_OP_MAP = {
    "__add__": "+",
    "__sub__": "-",
    "__mul__": "*",
    "__truediv__": "/",
    "__pow__": "**",
    "__eq__": "==",
    "__lt__": "<",
    "__le__": "<=",
    "__gt__": ">",
    "__ge__": ">=",
    "__ne__": "!=",
}


_UNARY_OP_MAP = {
    "__neg__": "-",
    "__invert__": "~",
    "__abs__": "abs",
}


def _contains_expression(repr_str: str) -> bool:
    """
    Checks whether or not a `repr_str` contains an expression. (Single unary expressions are excluded)
    """
    repr_str = re.sub(r"\s+", "", repr_str)
    repr_str = repr_str.replace("(", "")
    repr_str = repr_str.replace(")", "")

    symbols = re.findall(r"[\w']+", repr_str)
    non_symbols = re.findall(r"[^\w']+", repr_str)

    if len(non_symbols) == 0:
        return False

    if len(non_symbols) == 1 and len(symbols) == 1:
        return False

    return True


def _format_expression(f: tp.Any) -> str:
    """
    `f` could be either a single argument, or an expression of arguments. If it is the latter, wrap in parenthesis
    """
    repr_str = _repr(f)
    if _contains_expression(repr_str):
        return f"({repr_str})"
    return repr_str


def _repr(f: tp.Any) -> str:
    """Provide a string representation of the FN, recursively representing defined arguments."""

    def get_function_name(f: tp.Any) -> str:
        """Get a string representation of the callable, or its code if it is a lambda. In some cases, `f` may not be function, so just return a string."""
        if not isinstance(f, types.FunctionType) or not hasattr(f, "__name__"):
            return str(f)
        if f.__name__ == "<lambda>":
            # split on all white space, and rejoin with single space
            return " ".join(inspect.getsource(f).split())
        return f.__name__

    # find FunctionNode; using hasattr because of testing context issues
    if isinstance(f, FunctionNode):
        doc_f = get_function_name(f._doc_function)

        unary_op = _UNARY_OP_MAP.get(doc_f)
        binary_op = _BINARY_OP_MAP.get(doc_f)

        if unary_op:
            assert not f._doc_kwargs, "Unary FunctionNodes must not have doc_kwargs."
            assert (
                len(f._doc_args) == 1
            ), "Unary FunctionNodes must only have one doc_arg."

            if unary_op == "abs":
                arg = _repr(f._doc_args[0])
                return f"{unary_op}({arg})"

            arg = _format_expression(f._doc_args[0])
            return f"{unary_op}{arg}"

        if binary_op:
            assert not f._doc_kwargs, "Binary FunctionNodes must not have doc_kwargs."
            assert (
                len(f._doc_args) == 2
            ), "Binary FunctionNodes must only have two doc_args."

            left = _format_expression(f._doc_args[0])
            right = _format_expression(f._doc_args[1])

            return f"{left}{binary_op}{right}"

        if not f._doc_args and not f._doc_kwargs:
            return doc_f

        predecessor = ""
        sig_str = "("

        if f._doc_args:
            sig_str += ",".join((str(_repr(v)) for v in f._doc_args))

        if f._doc_kwargs:
            if f._doc_args:
                sig_str += ","

            for k, v in f._doc_kwargs.items():
                if k == PREDECESSOR_PN:
                    predecessor = _repr(v)
                else:
                    sig_str += (k + "=" + str(_repr(v))) + ","

        sig_str = sig_str.rstrip(",") + ")"

        if sig_str == "()":
            sig_str = doc_f
        else:
            sig_str = doc_f + sig_str

        if predecessor:
            sig_str = predecessor + " | " + sig_str
        return sig_str
    return get_function_name(f)


class FunctionNode:
    """
    A wrapper for a callable that can reside in an expression of numerous FunctionNodes, or be modified with unary or binary operators.
    """

    __slots__ = (
        "_function",
        "_doc_function",
        "_doc_args",
        "_doc_kwargs",
    )

    # ---------------------------------------------------------------------------
    def __init__(
        self: FN,
        function: tp.Any,
        *,
        doc_function: tp.Optional[tp.Callable] = None,
        doc_args: tp.Tuple[tp.Any, ...] = (),
        doc_kwargs: tp.Optional[tp.Dict[str, tp.Any]] = None,
    ) -> None:
        """
        Args:

            - ``function``: a callable or value. If given a value, will create a function that simply returns that value.
            - ``doc_function``: the function to display in the repr; will be set to ``function`` if not provided
            - ``doc_args``: the positional arguments to display in the repr
            - ``doc_kwargs``: the keyword arguments to display in the repr

        """
        # if a function node, re-wrap
        if isinstance(function, FunctionNode):
            for attr in self.__slots__:
                setattr(self, attr, getattr(function, attr))
        else:
            if callable(function):
                self._function = function
            else:
                # if not a callable, we upgrade a constant, non function value to be a function that returns that value
                self._function = lambda *args, **kwargs: function

            # if not supplied, doc_function is set to function
            self._doc_function = doc_function if doc_function else self._function
            self._doc_args = doc_args
            self._doc_kwargs = doc_kwargs

    @property
    def unwrap(self: FN) -> tp.Callable:
        """
        The doc_function should be set to the core function being wrapped, no matter the level of wrapping.
        """
        # if the stored function is using _pipe_kwarg_bind, need to go lower
        doc_func = self
        while hasattr(doc_func, "_doc_function"):
            doc_func = getattr(doc_func, "_doc_function")
        return doc_func

    def __call__(self: FN, *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        """
        Call the wrapped function with args and kwargs.
        """
        try:
            return self._function(*args, **kwargs)
        except Exception as e:
            raise _exception_with_cleaned_tb(e) from None

    def __str__(self: FN) -> str:
        return f"<FN: {_repr(self)}>"

    def __repr__(self: FN) -> str:
        return f"<FN: {_repr(self)}>"

    def partial(self: FN, *args: tp.Any, **kwargs: tp.Any) -> "FunctionNode":
        """
        Return a new FunctionNode with a partialed function with args and kwargs.
        """
        fn = FunctionNode(functools.partial(self._function, *args, **kwargs))

        for attr in self.__slots__:
            if not getattr(fn, attr):
                setattr(fn, attr, getattr(self, attr))

        return fn

    # -------------------------------------------------------------------------
    # Unary Operators

    @_wrap_unary
    def __neg__(self: FN) -> FN:
        """
        Return a new FunctionNode that when evaulated, will negate the result of ``self``
        """
        return lambda *args, **kwargs: self(*args, **kwargs) * -1  # type: ignore

    @_wrap_unary
    def __invert__(self: FN) -> FN:
        """
        Return a new FunctionNode that when evaulated, will invert the result of ``self``

        NOTE:
            This is generally expected to be a Boolean inversion, such as ~ (not) applied to a Numpy, Pandas, or Static-Frame objects.
        """
        return lambda *args, **kwargs: self(*args, **kwargs).__invert__()  # type: ignore

    @_wrap_unary
    def __abs__(self: FN) -> FN:
        """
        Return a new FunctionNode that when evaulated, will find the absolute value of the result of ``self``
        """
        return lambda *args, **kwargs: abs(self(*args, **kwargs))  # type: ignore

    # ---------------------------------------------------------------------------
    # all binary operators return a function; the _wrap_binary decorator then wraps this function in a FunctionNode definition and supplies appropriate doc args. Note both left and righ sides are wrapped in FNs to permit operations on constants

    @_wrap_binary
    def __add__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will add ``rhs`` to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) + self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __sub__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will subtract ``rhs`` to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) - self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __mul__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will multiply ``rhs`` by the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) * self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __truediv__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will divide ``rhs`` by the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) / self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __pow__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will divide raise the result of ``self`` by ``rhs``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) ** self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __radd__(self: FN, lhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will add ``rhs`` to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(lhs)(  # type: ignore
            *args, **kwargs
        ) + self.__class__(self)(*args, **kwargs)

    @_wrap_binary
    def __rsub__(self: FN, lhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will subtract ``rhs`` to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(lhs)(  # type: ignore
            *args, **kwargs
        ) - self.__class__(self)(*args, **kwargs)

    @_wrap_binary
    def __rmul__(self: FN, lhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will multiply ``rhs`` by the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(lhs)(  # type: ignore
            *args, **kwargs
        ) * self.__class__(self)(*args, **kwargs)

    @_wrap_binary
    def __rtruediv__(self: FN, lhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will divide ``rhs`` by the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(lhs)(  # type: ignore
            *args, **kwargs
        ) / self.__class__(self)(*args, **kwargs)

    # comparison operators, expected to return booleans
    @_wrap_binary
    def __eq__(self: FN, rhs: tp.Any) -> FN:  # type: ignore
        """
        Return a new FunctionNode will test if ``rhs``' equals the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) == self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __lt__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will test if ``rhs``' is less than the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) < self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __le__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will test if ``rhs``' is less than or equal to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) <= self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __gt__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will test if ``rhs``' is greater than the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) > self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __ge__(self: FN, rhs: tp.Any) -> FN:
        """
        Return a new FunctionNode will test if ``rhs``' is greater than or equal to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) >= self.__class__(rhs)(*args, **kwargs)

    @_wrap_binary
    def __ne__(self: FN, rhs: tp.Any) -> FN:  # type: ignore
        """
        Return a new FunctionNode will test if ``rhs``' is not equal to the result of ``self``
        """
        return lambda *args, **kwargs: self.__class__(self)(  # type: ignore
            *args, **kwargs
        ) != self.__class__(rhs)(*args, **kwargs)

    # ---------------------------------------------------------------------------
    # composition operators

    def __rshift__(self: FN, rhs: tp.Callable) -> FN:
        """
        Composes a new FunctionNode will call ``lhs`` first, and then feed its result into ``rhs``
        """
        return compose(rhs, self)

    def __rrshift__(self: FN, lhs: tp.Callable) -> FN:
        """
        Composes a new FunctionNode will call ``lhs`` first, and then feed its result into ``rhs``
        """
        return compose(self, lhs)

    def __lshift__(self: FN, rhs: tp.Callable) -> FN:
        """
        Composes a new FunctionNode will call ``rhs`` first, and then feed its result into ``lhs``
        """
        return compose(self, rhs)

    def __rlshift__(self: FN, lhs: tp.Callable) -> FN:
        """
        Composes a new FunctionNode will call ``rhs`` first, and then feed its result into ``lhs``
        """
        return compose(lhs, self)

    def __or__(self: FN, rhs: FN) -> FN:
        """Only implemented for PipeNode."""
        raise NotImplementedError()

    def __ror__(self: FN, lhs: FN) -> FN:
        """Only implemented for PipeNode."""
        raise NotImplementedError()


# -------------------------------------------------------------------------------
# PipeNode and utiltiies

# PipeNode kwargs
PREDECESSOR_RETURN = "predecessor_return"
PREDECESSOR_PN = "predecessor_pn"
PN_INPUT = "pn_input"
PN_INPUT_SET = frozenset((PN_INPUT,))
PREDECESSOR_PN_SET = frozenset((PREDECESSOR_PN,))
PIPE_NODE_KWARGS = frozenset((PREDECESSOR_RETURN, PREDECESSOR_PN, PN_INPUT))


class PipeNode(FunctionNode):
    """
    This encapsulates the node that will be used in a pipeline.

    It is not expected to be created directly, rather, through usage of ``pipe_node`` (and related) decorators.

    PipeNodes will be in (or move between) one of three states, depending on where it was created, or what the current state of pipeline evaluation is
    """

    class State(enum.Enum):
        """
        The current state of the PipeNode
        """

        FACTORY = "FACTORY"
        EXPRESSION = "EXPRESSION"
        PROCESS = "PROCESS"

    __slots__ = FunctionNode.__slots__ + (
        "_call_state",
        "_predecessor",
    )

    # ---------------------------------------------------------------------------
    def __init__(
        self: PN,
        function: tp.Any,
        *,
        doc_function: tp.Optional[tp.Callable] = None,
        doc_args: tp.Tuple[tp.Any, ...] = (),
        doc_kwargs: tp.Optional[tp.Dict[str, tp.Any]] = None,
        call_state: tp.Optional[State] = None,
        predecessor: tp.Optional[PN] = None,
    ):
        super().__init__(
            function=function,
            doc_function=doc_function,
            doc_args=doc_args,
            doc_kwargs=doc_kwargs,
        )
        self._call_state = call_state
        self._predecessor = predecessor

    def __str__(self: PN) -> str:
        if self._call_state is PipeNode.State.FACTORY:
            return f"<PNF: {_repr(self)}>"
        return f"<PN: {_repr(self)}>"

    def __repr__(self: PN) -> str:
        if self._call_state is PipeNode.State.FACTORY:
            return f"<PNF: {_repr(self)}>"
        return f"<PN: {_repr(self)}>"

    def partial(self: PN, *args: str, **kwargs: str) -> PN:
        """
        Partialing PipeNodes is prohibited. Use ``pipe_node_factory`` (and related) decorators to pass in expression-level arguments.
        """
        raise NotImplementedError()

    # ---------------------------------------------------------------------------
    # pipe node properties

    @property
    def call_state(self: PN) -> tp.Optional["State"]:
        """The current call state of the Node"""
        return self._call_state

    @property
    def predecessor(self: PN) -> tp.Optional[PN]:
        """
        The PipeNode preceeding this Node in a pipeline. Can be None
        """
        return self._predecessor

    # ---------------------------------------------------------------------------
    # composition operators

    def __rshift__(self: PN, rhs: tp.Callable) -> PN:
        """Only implemented for FunctionNode."""
        raise NotImplementedError()

    def __rrshift__(self: PN, lhs: tp.Callable) -> PN:
        """Only implemented for FunctionNode."""
        raise NotImplementedError()

    def __lshift__(self: PN, rhs: tp.Callable) -> PN:
        """Only implemented for FunctionNode."""
        raise NotImplementedError()

    def __rlshift__(self: PN, lhs: tp.Callable) -> PN:
        """Only implemented for FunctionNode."""
        raise NotImplementedError()

    def __or__(self: PN, rhs: PN) -> PN:
        """
        Invokes ``rhs``, passing in ``self`` as the kwarg ``PREDECESSOR_PN``.
        """
        return rhs(**{PREDECESSOR_PN: self})

    def __ror__(self: PN, lhs: PN) -> PN:
        """
        Invokes ``lhs``, passing in ``self`` as the kwarg ``PREDECESSOR_PN``.
        """
        return self(**{PREDECESSOR_PN: lhs})

    # ---------------------------------------------------------------------------

    def __getitem__(self: PN, pn_input: tp.Any) -> tp.Any:
        """
        Invokes ``self``, passing in ``pn_input`` as the kwarg ``PN_INPUT``.

        NOTE:
            - If ``None``, will evaluate self with a default ``PipeNodeInput`` instance
            - If user desires for the initial input to be literally ``None``, use ``(**{PN_INPUT: None})`` instead.
        """
        pn_input = pn_input if pn_input is not None else PipeNodeInput()
        return self(**{PN_INPUT: pn_input})

    def __call__(self: PN, *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        """
        Call the wrapped function with args and kwargs.
        """
        if self._call_state is PipeNode.State.FACTORY:
            try:
                return self._function(*args, **kwargs)
            except Exception as e:
                raise _exception_with_cleaned_tb(e) from None

        if args or set(kwargs) - PIPE_NODE_KWARGS != set():
            raise ValueError(
                "Cannot call a PipeNode with args or non-pipeline kwargs! Perhaps you meant to use @pipe_node_factory?"
            )

        try:
            return self._function(**kwargs)
        except Exception as e:
            raise _exception_with_cleaned_tb(e) from None


# -------------------------------------------------------------------------------
# decorator utilities


def _broadcast(
    *,
    factory_args: tp.Tuple[tp.Any, ...],
    factory_kwargs: tp.Dict[str, tp.Any],
    processing_args: tp.Tuple[tp.Any, ...] = (),
    processing_kwargs: tp.Dict[str, tp.Any],
) -> tp.Tuple[tp.Tuple[tp.Any, ...], tp.Dict[str, tp.Any]]:
    """
    Factory args/kwargs are those given to pipe_node_factory at the expression level.
    Processing args/kwargs are those given as the initial input, and used to call all processing functions.

    After calling factory args with processing args, the result is used as core_callable args
    """
    core_callable_args = tuple(
        arg(*processing_args, **processing_kwargs) if isinstance(arg, PipeNode) else arg
        for arg in factory_args
    )

    core_callable_kwargs = {
        kw: arg(*processing_args, **processing_kwargs)
        if isinstance(arg, PipeNode)
        else arg
        for kw, arg in factory_kwargs.items()
    }

    return core_callable_args, core_callable_kwargs


def _core_logger(core_callable: tp.Callable) -> tp.Callable:
    """
    A decorator to provide output on the execution of each core callable call. Alternative decorators can be used to partial pipe_node_factory and pipe_node.
    """

    @functools.wraps(core_callable)
    def wrapped(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        print("|", str(core_callable), file=sys.stderr)
        return core_callable(*args, **kwargs)

    return wrapped


def _has_key_positions(*key_positions: KeyPostion) -> bool:
    """
    Returns whether or not key_positions is a list of key positions, or if it is just a single callable
    """
    return not bool(len(key_positions) == 1 and callable(key_positions[0]))


def _is_unbound_self_method(
    core_callable: tp.Union[classmethod, staticmethod, tp.Callable],
    *,
    self_keyword: str,
) -> bool:
    """
    Inspects a given callable to determine if it's both unbound, and the first argument in its signature is ``self_keyword``
    """
    if isinstance(core_callable, types.MethodType):
        return False

    if isinstance(core_callable, (staticmethod, classmethod)):
        return False

    if isinstance(core_callable, functools.partial):
        return False

    argspec = inspect.getfullargspec(core_callable)
    return bool(argspec.args and argspec.args[0] == self_keyword)


def _pipe_kwarg_bind(
    *key_positions: KeyPostion,
) -> tp.Callable[[tp.Callable], tp.Callable]:
    """
    Binds a specific PN labels wrapped up in **kwargs to the first n positional arguments of the core callable
    """

    def decorator(core_callable: tp.Callable) -> tp.Callable:
        @functools.wraps(core_callable)
        def wrapped(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
            target_args = [kwargs.pop(key) for key in key_positions]  # type: ignore
            target_kwargs = {
                k: v for k, v in kwargs.items() if k not in PIPE_NODE_KWARGS
            }
            return core_callable(*target_args, *args, **target_kwargs)

        return wrapped

    return decorator


class PipeNodeDescriptor:  # pylint: disable=too-few-public-methods
    """
    Wraps up ``pipe_node``/``pipe_node_factory`` behavior in a descriptor, where it will bind instance and owner to the core_callable, and then pass it along the pipeline
    """

    __slots__ = (
        "core_callable",
        "core_handler",
        "key_positions",
    )

    def __init__(
        self: PipeNodeDescriptorT,
        core_callable: tp.Callable,
        core_handler: HandlerT,
        key_positions: tp.Optional[tp.Tuple[KeyPostion, ...]] = None,
    ) -> None:
        self.core_callable = core_callable
        self.core_handler = core_handler
        self.key_positions = key_positions

    def __get__(
        self: PipeNodeDescriptorT,
        instance: tp.Any,
        owner: tp.Any,
    ) -> tp.Callable:
        """
        Returns a callable that will be bound to the instance/owner, and then passed along the pipeline.
        """
        core_callable: tp.Callable = self.core_callable.__get__(instance, owner)  # type: ignore
        if self.key_positions is not None:
            core_callable = _pipe_kwarg_bind(*self.key_positions)(core_callable)
        return self.core_handler(core_callable)


def _handle_descriptors_and_key_positions(
    *key_positions: KeyPostion,
    core_handler: HandlerT,
    self_keyword: str,
) -> tp.Union[
    PipeNodeDescriptor,
    HandlerT,
    tp.Callable[[tp.Callable], tp.Union[PipeNodeDescriptor, HandlerT]],
]:
    """
    We can return either a callable or a ``PipeNodeDescriptor``, OR, a decorator that when called,
    will return either a callable or a ``PipeNodeDescriptor``.
    """
    has_key_positions = _has_key_positions(*key_positions)

    # See if decorator was given no arguments, and received the core_callable directly.
    if not has_key_positions:
        final_callable = key_positions[0]
        assert callable(final_callable), (type(final_callable), final_callable)

        if _is_unbound_self_method(final_callable, self_keyword=self_keyword):
            return PipeNodeDescriptor(final_callable, core_handler)

        return core_handler(final_callable)

    def decorator_wrapper(
        core_callable: tp.Callable,
    ) -> tp.Union[PipeNodeDescriptor, HandlerT]:
        if _is_unbound_self_method(core_callable, self_keyword=self_keyword):
            return PipeNodeDescriptor(core_callable, core_handler, key_positions)

        final_callable = _pipe_kwarg_bind(*key_positions)(core_callable)
        return core_handler(final_callable)

    return decorator_wrapper


def _descriptor_factory(
    *key_positions: KeyPostion,
    decorator: tp.Callable,
    core_decorator: HandlerT,
    emulator: tp.Any,
) -> tp.Any:

    has_key_positions = _has_key_positions(*key_positions)

    class Descriptor:  # pylint: disable=too-few-public-methods
        def __init__(self, func: tp.Callable) -> None:
            self._func = func

        def __get__(self, instance: tp.Any, owner: tp.Any) -> tp.Callable:

            assert emulator is staticmethod or emulator is classmethod, emulator

            # Prefer this to partialing for prettier func reprs
            @functools.wraps(self._func)
            def func(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
                if emulator is staticmethod:
                    return self._func(*args, **kwargs)
                return self._func(owner, *args, **kwargs)

            if has_key_positions:
                func = _pipe_kwarg_bind(*key_positions)(func)

            return decorator(func, core_decorator=core_decorator)

    if not has_key_positions:
        return Descriptor(key_positions[0])  # type: ignore

    return Descriptor


# -------------------------------------------------------------------------------
# decorators


def pipe_node_factory(
    *key_positions: KeyPostion,
    core_decorator: HandlerT = _core_logger,
    self_keyword: str = "self",
) -> tp.Union[tp.Callable, tp.Callable[[tp.Any], PipeNode]]:
    """
    Decorates a function to become a pipe node factory, that when given *expression-level* arguments, will return a ``PipeNode``

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    **Examples**:

    >>> @pipe_node_factory
    >>> def func(a, b, **kwargs):
    >>>     pass
    >>> ...
    >>> func(1, 2) # This is now a PipeNode!

    >>> @pipe_node_factory()
    >>> def func(*, a, b):
    >>>     pass
    >>> ...
    >>> func(a=1, b=2) # This is now a PipeNode!

    >>> @pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
    >>> def func(pn_input, previous_value, a, *, b):
    >>>     # pn_input will be given the PN_INPUT from the pipeline
    >>>     # prev will be given the PREDECESSOR_RETURN from the pipeline
    >>>     pass
    >>> ...
    >>> func(1, b=2) # This is now a PipeNode!

    >>> class Example:
    >>>     @pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
    >>>     def method(self, pn_input, previous_value, a, *, b):
    >>>         pass
    >>> ...
    >>> Example().method(1, b=2) # This is now a PipeNode!

    >>> from functools import partial
    >>> español_pipe_node_factory = partial(pipe_node_factory, self_keyword="uno_mismo")
    >>> ...
    >>> class Ejemplo:
    >>>     @español_pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
    >>>     def método(uno_mismo, pn_input, valor_anterior, a, *, b):
    >>>         pass
    >>> ...
    >>> Ejemplo().método(1, b=2) # Esto ahora es un PipeNode!

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
        - ``self_keyword``: which keyword to look for when decorating instance methods.
    """

    def build_factory(core_callable: tp.Callable) -> PipeNode:
        decorated_core_callable = core_decorator(core_callable)

        def factory_f(*f_args: tp.Any, **f_kwargs: tp.Any) -> PipeNode:
            """This is the function returned by the decorator, used to create the PipeNode that resides in expressions after being called with arguments.

            f_args and f_kwargs are passed to the core_callable; if f_args or f_kwargs are PipeNode instances, they will be called with the processing args and kwargs (including PN_INPUT), either from process_f or (if innermost) from expression args.
            """
            if set(f_kwargs).intersection(PIPE_NODE_KWARGS) != set():
                raise ValueError(
                    f"Either you put a factory in a pipeline (i.e. not a pipe node), or your factory was given a reserved pipeline kwarg {tuple(PIPE_NODE_KWARGS)}."
                )

            def expression_f(**e_kwargs: tp.Any) -> PipeNode:
                """This is the PipeNode that resides in expressions prior to `|` operator evalation.
                When called with `|`, the predecessor is passed is in e_kwargs as PREDECESSOR_PN. In this usage the e_args will always be empty.

                When in the innermost position, expression_f is never called with `|` but with the PN_INPUT; this sitation is identified and the core_callable is called immediately.

                e_args will only be used as an innermost call.
                """
                kwargset = set(e_kwargs)
                if kwargset not in (PN_INPUT_SET, PREDECESSOR_PN_SET):
                    raise ValueError(
                        f"Expected to be called with either {PN_INPUT} or {PREDECESSOR_PN} only, not: {kwargset}"
                    )

                # identify innermost condition as when the expression level kwargs consists only of PN_INPUT
                if kwargset == PN_INPUT_SET:
                    # as this is innermost, processing args (i.e., PipeNodeInput) are given here at the expression level (as no Pipe operator has been used to call the innermost)
                    core_callable_args, core_callable_kwargs = _broadcast(
                        factory_args=f_args,
                        factory_kwargs=f_kwargs,
                        processing_kwargs=e_kwargs,
                    )  # not p_kwargs

                    # pack PipeNode protocol kwargs; when used as innermost, a core_callable can only expect to have a PN_INPUT
                    core_callable_kwargs[PN_INPUT] = e_kwargs[PN_INPUT]

                    return decorated_core_callable(
                        *core_callable_args, **core_callable_kwargs
                    )

                predecessor_pn = e_kwargs[PREDECESSOR_PN]

                def process_f(*p_args: tp.Any, **p_kwargs: tp.Any) -> tp.Any:
                    # call the predecssor PipeNode (here a process_f) with these processing args; these are always the args given as the initial input to the innermost function, generally a PipeNodeInput
                    predecessor_return = predecessor_pn(*p_args, **p_kwargs)

                    core_callable_args, core_callable_kwargs = _broadcast(
                        factory_args=f_args,
                        factory_kwargs=f_kwargs,
                        processing_args=p_args,
                        processing_kwargs=p_kwargs,
                    )

                    # pack PipeNode protocol kwargs
                    core_callable_kwargs[PN_INPUT] = p_kwargs[PN_INPUT]
                    core_callable_kwargs[PREDECESSOR_PN] = predecessor_pn
                    core_callable_kwargs[PREDECESSOR_RETURN] = predecessor_return

                    return decorated_core_callable(
                        *core_callable_args, **core_callable_kwargs
                    )

                assert (
                    set(e_kwargs).intersection(set(f_kwargs)) == set()
                )  # This is in impossible state

                # we must return a PipeNode here, as this is the final thing returned and might be passed on to another series func
                return PipeNode(
                    process_f,
                    doc_function=core_callable,
                    doc_args=f_args,
                    doc_kwargs={**e_kwargs, **f_kwargs},
                    call_state=PipeNode.State.PROCESS,
                    predecessor=predecessor_pn,
                )

            return PipeNode(
                expression_f,
                doc_function=core_callable,
                doc_args=f_args,
                doc_kwargs=f_kwargs,
                call_state=PipeNode.State.EXPRESSION,
            )

        # return a function node so as to make doc_function available in test
        return PipeNode(
            factory_f,
            doc_function=core_callable,
            call_state=PipeNode.State.FACTORY,
        )

    return _handle_descriptors_and_key_positions(  # type: ignore
        *key_positions,
        core_handler=build_factory,
        self_keyword=self_keyword,
    )


def pipe_node(
    *key_positions: KeyPostion,
    core_decorator: HandlerT = _core_logger,
    self_keyword: str = "self",
) -> tp.Union[tp.Callable, PipeNode]:
    """
    Decorates a function to become a ``PipeNode`` that takes no expression-level args.

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    **Examples**:

    >>> @pipe_node
    >>> def func(**kwargs):
    >>>     pass

    >>> @pipe_node()
    >>> def func():
    >>>     pass

    >>> @pipe_node(PN_INPUT)
    >>> def func(pn_input):
    >>>     pass

    >>> class Example:
    >>>     @pipe_node(PN_INPUT)
    >>>     def method(self, pn_input):
    >>>         pass

    >>> from functools import partial
    >>> español_pipe_node = partial(pipe_node, self_keyword="uno_mismo")
    >>> ...
    >>> class Ejemplo:
    >>>     @español_pipe_node(PN_INPUT)
    >>>     def método(uno_mismo, pn_input):
    >>>         pass

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
        - ``self_keyword``: which keyword to look for when decorating instance methods.
    """

    def create_factory_and_call_once(core_callable: tp.Callable) -> PipeNode:
        # Create a factory and call it once with no args to get an expresion-level function
        pnf = pipe_node_factory(core_callable, core_decorator=core_decorator)

        if not callable(pnf):
            raise ValueError(f"{core_callable.__qualname__} requires an instance")

        return pipe_node_factory(core_callable, core_decorator=core_decorator)()  # type: ignore

    return _handle_descriptors_and_key_positions(  # type: ignore
        *key_positions,
        core_handler=create_factory_and_call_once,
        self_keyword=self_keyword,
    )


def classmethod_pipe_node_factory(
    *key_positions: KeyPostion, core_decorator: HandlerT = _core_logger
) -> tp.Callable:
    """
    Decorates a function to become a classmethod pipe node factory, that when given *expression-level* arguments, will return a ``PipeNode``

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    This is a convenience method, that is the mental equivalent to this pseudo-code:

    >>> @classmethod
    >>> @pipe_node_factory(...)
    >>> def func(...)

    **Examples**:

    >>> @classmethod_pipe_node_factory
    >>> def func(cls, a, b, **kwargs):
    >>>     pass
    >>> ...
    >>> SomeClass.func(1, 2) # This is now a PipeNode!

    >>> @classmethod_pipe_node_factory()
    >>> def func(cls, *, a, b):
    >>>     pass
    >>> ...
    >>> SomeClass.func(a=1, b=2) # This is now a PipeNode!

    >>> @classmethod_pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
    >>> def func(cls, pn_input, previous_value, a, *, b):
    >>>     # ``pn_input`` will be given the PN_INPUT from the pipeline
    >>>     # ``previous_value`` will be given the PREDECESSOR_RETURN from the pipeline
    >>>     pass
    >>> ...
    >>> SomeClass.func(1, b=2) # This is now a PipeNode!

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
    """
    return _descriptor_factory(
        *key_positions,
        decorator=pipe_node_factory,
        core_decorator=core_decorator,
        emulator=classmethod,
    )


def classmethod_pipe_node(
    *key_positions: KeyPostion,
    core_decorator: HandlerT = _core_logger,
) -> tp.Union[tp.Callable, PipeNode]:
    """
    Decorates a function to become a classmethod ``PipeNode`` that takes no expression-level args.

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    This is a convenience method, that is the mental equivalent to this pseudo-code:

    >>> @classmethod
    >>> @pipe_node(...)
    >>> def func(...)

    **Examples**:

    >>> @classmethod_pipe_node
    >>> def func(cls, **kwargs):
    >>>     pass

    >>> @classmethod_pipe_node()
    >>> def func(cls):
    >>>     pass

    >>> @classmethod_pipe_node(PN_INPUT)
    >>> def func(cls, pn_input):
    >>>     pass

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
    """
    return _descriptor_factory(
        *key_positions,
        decorator=pipe_node,
        core_decorator=core_decorator,
        emulator=classmethod,
    )


def staticmethod_pipe_node_factory(
    *key_positions: KeyPostion, core_decorator: HandlerT = _core_logger
) -> tp.Callable:
    """
    Decorates a function to become a staticmethod pipe node factory, that when given *expression-level* arguments, will return a ``PipeNode``

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    This is a convenience method, that is the mental equivalent to this pseudo-code:

    >>> @staticmethod
    >>> @pipe_node_factory(...)
    >>> def func(...)

    **Examples**:

    >>> @staticmethod_pipe_node_factory
    >>> def func(a, b, **kwargs):
    >>>     pass
    >>> ...
    >>> SomeClass.func(1, 2) # This is now a PipeNode!

    >>> @staticmethod_pipe_node_factory()
    >>> def func(*, a, b):
    >>>     pass
    >>> ...
    >>> SomeClass.func(a=1, b=2) # This is now a PipeNode!

    >>> @staticmethod_pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
    >>> def func(pn_input, previous_value, a, *, b):
    >>>     # ``pn_input`` will be given the PN_INPUT from the pipeline
    >>>     # ``previous_value`` will be given the PREDECESSOR_RETURN from the pipeline
    >>>     pass
    >>> ...
    >>> SomeClass.func(1, b=2) # This is now a PipeNode!

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
    """
    return _descriptor_factory(
        *key_positions,
        decorator=pipe_node_factory,
        core_decorator=core_decorator,
        emulator=staticmethod,
    )


def staticmethod_pipe_node(
    *key_positions: KeyPostion,
    core_decorator: HandlerT = _core_logger,
) -> tp.Union[tp.Callable, PipeNode]:
    """
    Decorates a function to become a staticmethod ``PipeNode`` that takes no expression-level args.

    This can either be used as a decorator, or a decorator factory, similar to ``functools.lru_cache``.

    This is a convenience method, that is the mental equivalent to this pseudo-code:

    >>> @staticmethod
    >>> @pipe_node(...)
    >>> def func(...)

    **Examples**:

    >>> @staticmethod_pipe_node
    >>> def func(**kwargs):
    >>>     pass

    >>> @staticmethod_pipe_node()
    >>> def func():
    >>>     pass

    >>> @staticmethod_pipe_node(PN_INPUT)
    >>> def func(pn_input):
    >>>     pass

    Args:
        - ``key_positions``: either a single callable, or a list of keywords that will be positionally bound to the decorated function.
        - ``core_decorator``: a decorator that will be applied to the core_callable. This is typically a logger. By default, it will print to stdout.
    """
    return _descriptor_factory(
        *key_positions,
        decorator=pipe_node,
        core_decorator=core_decorator,
        emulator=staticmethod,
    )


# -------------------------------------------------------------------------------
# PipeNodeInput


class PipeNodeInput:
    """
    PipeNode input to support store and recall; subclassable to expose other attributes and parameters.
    """

    def __init__(self: PNI) -> None:
        self._store: tp.Dict[str, tp.Any] = {}

    def store(self: PNI, key: str, value: tp.Any) -> None:
        """Store ``key`` and ``value`` in the underlying store."""
        if key in self._store:
            raise KeyError("cannot store the same key", key)
        self._store[key] = value

    def recall(self: PNI, key: str) -> tp.Any:
        """Recall ``key`` from the underlying store. Can raise an ``KeyError``"""
        return self._store[key]

    @property
    def store_items(self: PNI) -> tp.ItemsView[str, tp.Any]:
        """Return an items view of the underlying store."""
        return self._store.items()


# -------------------------------------------------------------------------------
# Utility PipeNodes


@pipe_node_factory(PN_INPUT, PREDECESSOR_RETURN)
def store(pn_input: PipeNodeInput, previous_value: tp.Any, label: str) -> tp.Any:
    """
    Store ``previous_value`` (the value returned from the previous ``PipeNode``) to ``pn_input`` under ``label``. Forward ``previous_value``.
    """
    pn_input.store(label, previous_value)
    return previous_value


@pipe_node_factory(PN_INPUT)
def recall(pn_input: PipeNodeInput, label: str) -> tp.Any:
    """
    Recall ``label`` from ``pn_input```` and return it. Can raise an ``KeyError``.
    """
    return pn_input.recall(label)


@pipe_node_factory()
def call(*pns: PipeNode) -> tp.Any:
    """
    Since ``pns`` are all ``PipeNodes``, they will all be evaluated before passed in as values to this function.

    This is called broadcasting.

    After they have all been evaluated, return the last result.
    """
    return pns[-1]  # the last result is returned



# To use a consistent encoding
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import setup

# https://packaging.python.org/distributing/
# to deploy:
# rm -r dist
# python setup.py sdist
# python setup.py bdist_wheel
# twine upload dist/*

ROOT_DIR_FP = path.abspath(path.dirname(__file__))

def get_long_description() -> str:
    with open(path.join(ROOT_DIR_FP, "README.rst"), encoding="utf-8") as f:
        return f.read()

def get_version() -> str:
    with open(path.join(ROOT_DIR_FP, 'function_pipe', '__init__.py'),
            encoding='utf-8') as f:
        for l in f:
            if l.startswith('__version__'):
                if '#' in l:
                    l = l.split('#')[0].strip()
                return l.split('=')[-1].strip()[1:-1]
    raise ValueError('__version__ not found!')

setup(
    name="function-pipe",
    version=get_version(),

    description="Tools for extended function composition and pipelines",
    long_description=get_long_description(),

    url="https://github.com/InvestmentSystems/function-pipe",
    author="Christopher Ariza, Charles Burkland",
    license="MIT",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        ],

    keywords="functionnode pipenode composition pipeline pipe",
    #py_modules=["function_pipe"], # no .py!
    packages=["function_pipe",
        "function_pipe.core",
        "function_pipe.test",
        ],
)
